#!/usr/bin/perl
# Convert an XM module to a SNES MSS song
# The output is a binary image that needs to get loaded to the SPC700 RAM
# starting at 0x200. Once it is loaded you then tell the SPC to jump to the
# MSS player at 0x400. Then you send a command to the SPC to tell the MSS
# player to start playing the tune, whose address is printed by this script.

open(XM,"<level_2.xm");

# Read XM header
read(XM,$magic,17);
if($magic ne "Extended Module: ") {
	die "Not an XM";
}
seek XM,0x3a,0;
$version=readword();
if($version!=0x0104) {
	die "XM version is $version wanted 0x0104";
}
$headersize=readlong();
$songlength=readword();
$restartpos=readword();
$channels=readword();
if($channels>8) {
	die "#channels is $channels wanted <=8";
}
$patterns=readword();
$instruments=readword();
$freqflag=readword();
$default_tempo=readword();
$default_bpm=readword();
for($i=0;$i<256;$i++) {
	push @pattern_order, readbyte();
}

# Read patterns
print STDERR "headersize=$headersize ";
print STDERR "position=".tell(XM)." ";
print STDERR "headersize+3c=".($headersize+0x3c)."\n";
print STDERR "Reading patterns\n";
seek XM,$headersize+0x3C,0;
for($i=0;$i<$patterns;$i++) {
	print STDERR "pattern $i ";
	$pattern[$i]->{headersize}=readlong();
	$pattern[$i]->{packtype}=readbyte();
	$pattern[$i]->{rows}=readword();
	$pattern[$i]->{packedsize}=readword();
	print STDERR "headersize=$pattern[$i]->{headersize} packedsize=$pattern[$i]->{packedsize} rows=$pattern[$i]->{rows} ";
	if($pattern[$i]->{packedsize}>0) {
		read(XM,$pattern[$i]->{data},$pattern[$i]->{packedsize})
	}
}
print STDERR "\n";

# Read instruments
print STDERR "Reading instruments\n";
for($i=0;$i<$instruments;$i++) {
	# instrument header
	$inst[$i]->{headerstart}=tell(XM);
	$inst[$i]->{headersize}=readlong();
	read(XM,$inst[$i]->{name},22);
	$inst[$i]->{type}=readbyte();
	$inst[$i]->{samples}=readword();
	if($inst[$i]->{samples}>0) {
		# instrument extra-header
		$inst[$i]->{sampheadersize}=readlong();
		for($j=0;$j<96;$j++) {
			$inst[$i]->{samplenum}->[$j]=readbyte();
		}
		read(XM,$inst[$i]->{venv},48);
		read(XM,$inst[$i]->{penv},48);
		$inst[$i]->{vpoints}=readbyte();
		$inst[$i]->{ppoints}=readbyte();
		$inst[$i]->{vsustain}=readbyte();
		$inst[$i]->{vloops}=readbyte();
		$inst[$i]->{vloope}=readbyte();
		$inst[$i]->{psustain}=readbyte();
		$inst[$i]->{ploops}=readbyte();
		$inst[$i]->{ploope}=readbyte();
		$inst[$i]->{vtype}=readbyte();
		$inst[$i]->{ptype}=readbyte();
		$inst[$i]->{vibratot}=readbyte();
		$inst[$i]->{vibratos}=readbyte();
		$inst[$i]->{vibratod}=readbyte();
		$inst[$i]->{vibrator}=readbyte();
		$inst[$i]->{wolf}=readbyte();
		seek XM,$inst[$i]->{headerstart}+$inst[$i]->{headersize},0;
		# sample headers
		# (yes -- all sample headers, then all sample data)
		for($j=0;$j<$inst[$i]->{samples};$j++) {
			$inst[$i]->{samp}->[$j]->{length}=readlong();
			$inst[$i]->{samp}->[$j]->{loopstart}=readlong();
			$inst[$i]->{samp}->[$j]->{looplength}=readlong();
			$inst[$i]->{samp}->[$j]->{volume}=readbyte();
			$inst[$i]->{samp}->[$j]->{finetune}=readbyte();
			if($inst[$i]->{samp}->[$j]->{finetune}&0x80) {
				$inst[$i]->{samp}->[$j]->{finetune}-=256;
				#$inst[$i]->{samp}->[$j]->{finetune}=127-$inst[$i]->{samp}->[$j]->{finetune};
			}
			$inst[$i]->{samp}->[$j]->{flags}=readbyte();
			$inst[$i]->{samp}->[$j]->{panning}=readbyte();
			$inst[$i]->{samp}->[$j]->{relnote}=readbyte();
			if($inst[$i]->{samp}->[$j]->{relnote}&0x80) {
				$inst[$i]->{samp}->[$j]->{relnote}-=256;
				#$inst[$i]->{samp}->[$j]->{relnote}=127-$inst[$i]->{samp}->[$j]->{relnote};
			}
			$inst[$i]->{samp}->[$j]->{reserved}=readbyte();
			read(XM,$inst[$i]->{samp}->[$j]->{name},22);
			print "sample $j for instrument $i ($inst[$i]->{samp}->[$j]->{name}]\n";
			print "relnote=$inst[$i]->{samp}->[$j]->{relnote}, finetune=$inst[$i]->{samp}->[$j]->{finetune}\n";
		}
		# sample data
		for($j=0;$j<$inst[$i]->{samples};$j++) {
			#printf "Reading instrument $i sample $j data, %d bytes\n",$inst[$i]->{samp}->[$j]->{length};
			read(XM,$inst[$i]->{samp}->[$j]->{data},$inst[$i]->{samp}->[$j]->{length});
			#printf "read %d bytes\n",length($inst[$i]->{samp}->[$j]->{data});
			#printf "offset in input file is now %d\n",tell XM;
			#print $inst[$i]->{samp}->[$j]->{data};
			#for($xxx=0;$xxx<length($inst[$i]->{samp}->[$j]->{data});$xxx++) {
			#	printf("%02x ",ord(substr($inst[$i]->{samp}->[$j]->{data},$xxx,1)));
			#}
		}
	}
}

if(1) {
	print "$patterns patterns, $instruments instruments\n";
	for($i=0;$i<$patterns;$i++) {
		print "pattern $i: $pattern[$i]->{rows} rows\n";
	}
	for($i=0;$i<$instruments;$i++) {
		print "instrument $i: $inst[$i]->{samples} samples\n";
	}
}

close XM;

# now encode the samples.
# the apu can handle 256 samples, but mss's memory map only allows for 128
# each of our instruments above has 1 or more samples
# we need to allocate an apu sample number (0..127) for each of the
# instrument-sample pairs. then while playing a pattern, it calls out the
# apu sample number; we look up the apu sample number from the instrument-sample
# pair when we go to convert the pattern.

open(SPC,">level_2.apu");

$apusample=0;
$spcaddr=0x600;
print "\$instruments=$instruments\n";
for($i=0;$i<$instruments;$i++) {
	print "inst[$i]->{samples}=$inst[$i]->{samples}\n";
	for($j=0;$j<$inst[$i]->{samples};$j++) {
		$inst[$i]->{samp}->[$j]->{apusample}=$apusample;
		# write the source directory entry
		print "inst $i samp $j is srcd $apusample, at $spcaddr\n";
		seek SPC,$apusample*4,0;
		print SPC chr($spcaddr&0xff); # start address low
		print SPC chr($spcaddr>>8); # start address high
		print SPC chr($spcaddr&0xff); # loop start address low
		print SPC chr($spcaddr>>8); # loop start address high
		# write the sample
		writesample($i,$j);
		$apusample++;
	}
}

# ok, now write the patterns
#seek SPC,$spcaddr-0x200,0;
for($i=0;$i<$patterns;$i++) {
	print "pattern $i is at $spcaddr\n";
	$pattern[$i]->{spcaddr}=$spcaddr;
	$data=$pattern[$i]->{data};
	$dataptr=0;
	$last_instrument=-1;
	for($r=0;$r<$pattern[$i]->{rows};$r++) {
		print STDERR "p$i r$r: ";
		$key_on_mask=0;
		$key_off_mask=0;
		for($chan=0;$chan<$channels;$chan++) {
			#print "row $r channel $chan\n";
			%notedata=readnote();
			printf STDERR "%02x %02x %02x %01x%02x  ",$notedata{note},$notedata{instrument},$notedata{volume},$notedata{effect_type},$notedata{effect_parm};
			if($notedata{instrument}!=0) {
				$new_instrument=$notedata{instrument};
			} else {
				$new_instrument=$chaninfo->[$chan]->{last_instrument};
			}
			$chaninfo->[$chan]->{last_instrument}=$new_instrument;
			# i'm assuming no one will change instruments mid-note
			# without starting a new note...
			if($notedata{note}==0) {
				# sustain last note
			} elsif($notedata{note}==97) {
				# key off
				print "channel $chan key off\n";
				#$keys=chr(1<<$chan); # voice nums to key off
				#writespc("\x09$keys"); # key off
				#$key_off_mask|=1<<$chan;
# XXX. we're not supposed to send bunches of KON/KOFF
# for every row. you have to allow 64.2us (=32? machine
# cycles) between successive writes. so, we should
# instead save up all KON/KOFF and write only one of
# each per row.
			#} elsif($notedata{note}==3) {
			} else {
				# new note
				print "channel $chan new note, instrument=$new_instrument, note=$notedata{note}\n";
				$keys=chr(1<<$chan); # voice nums to key off
				$key_off_mask|=$keys;
#				writespc("\x09$keys"); # key off (old note)
				# set SRCN (sample number)
				$srcn=chr(figure_srcn($new_instrument,$notedata{note}));
				if($srcn ne $chaninfo->[$chan]->{last_srcn}) {
					$reg=chr(0x04+($chan<<4)); # voice $chan SRCN
					printf("setting srcn reg %x=%x\n",ord($reg),ord($srcn));
					writespc("\x02$reg$srcn"); # set SRCN
					$chaninfo->[$chan]->{last_srcn}=$srcn;
				}
				# set pitch
				$pitch=figure_pitch($new_instrument,$notedata{note});
				if($pitch ne $chaninfo->[$chan]->{last_pitch}) {
					$cmd=chr(0x10+$chan); # set pitch voice $chan
					printf("setting pitch cmd %x=%x\n",ord($cmd),unpack("v",$pitch));
					writespc("$cmd$pitch"); # set pitch
					$chaninfo->[$chan]->{last_pitch}=$pitch;
				}
				# key on
#if($chan!=0) {
				#printf("keying on, mask=%x\n",ord($keys));
				#writespc("\x08$keys"); # key on (new note)
#}
				$key_on_mask|=1<<$chan;
			}
			# currently we IGNORE volume and effects stuff
			print "warning! volume=$notedata{volume}\n" if($notedata{volume}!=0);
			print "warning! effect_type=$notedata{effect_type}\n" if($notedata{effect_type}!=0);
			print "warning! effect_parm=$notedata{effect_parm}\n" if($notedata{effect_parm}!=0);
		}
		if($key_off_mask!=0) {
			printf("keying off, mask=%x\n",$key_off_mask);
			writespc("\x09".chr($key_off_mask));
		}
		if($key_on_mask!=0) {
			printf("keying on, mask=%x\n",$key_on_mask);
			writespc("\x08".chr($key_on_mask));
		}
		$ticks=chr(figure_ticks());
		printf("delay for %d/128ths sec\n",ord($ticks));
		writespc("\x01$ticks");
		print STDERR "\n";
	}
	writespc("\x06"); # RTS
}

# write out the code that initializes the fu
$mss_initializer_addr=$spcaddr;
# init code from SongInit.ASM
writespc("\x02\x02\x00"); # crap for voice 0
writespc("\x02\x03\x10");
writespc("\x02\x04\x00");
writespc("\x02\x05\xfa");
writespc("\x02\x06\xe4");
writespc("\x02\x15\xfa"); # uh... let's copy the ADSR settings to the other
writespc("\x02\x16\xe4"); # channels too.. just in case
writespc("\x02\x25\xfa");
writespc("\x02\x26\xe4");
writespc("\x02\x35\xfa");
writespc("\x02\x36\xe4");
writespc("\x02\x45\xfa");
writespc("\x02\x46\xe4");
writespc("\x02\x55\xfa");
writespc("\x02\x56\xe4");
writespc("\x02\x65\xfa");
writespc("\x02\x66\xe4");
writespc("\x02\x75\xfa");
writespc("\x02\x76\xe4");
# init code from mss.txt
writespc("\x02\x07\xa0"); # gain
writespc("\x02\x17\xa0"); # gain
writespc("\x02\x27\xa0"); # gain
writespc("\x02\x37\xa0"); # gain
writespc("\x02\x47\xa0"); # gain
writespc("\x02\x57\xa0"); # gain
writespc("\x02\x67\xa0"); # gain
writespc("\x02\x77\xa0"); # gain
writespc("\x02\x07\x7f\x02\x17\x7f\x02\x27\x7f\x02\x37\x7f\x02\x47\x7f\x02\x57\x7f"); # play loud
writespc("\x02\x0c\x40"); # main volume left
writespc("\x02\x1c\x40"); # main volume right
writespc("\x02\x0c\x7f\x02\x0c\x7f"); # play loud
writespc("\x02\x2c\x00"); # echo volume left - off
writespc("\x02\x3c\x00"); # echo volume right
writespc("\x02\x4d\x00"); # echo on/off
writespc("\x02\x3d\x00"); # noise on/off
writespc("\x02\x2d\x00"); # pitch modulation
#writespc("\x02\x6c\x00"); # misc settings
writespc("\x02\x6c\x20"); # misc settings
writespc("\x02\x0d\x5a"); # echo feedback
writespc("\x02\x6d\xc1"); # echo region start address - wtf? 0xc100 not mapped!
writespc("\x02\x0f\x7f"); # echo filter coeff 0 (clean filter)
writespc("\x02\x00\x1f"); # set l/r volumes
writespc("\x02\x01\x1f"); # set l/r volumes
writespc("\x02\x10\x1f"); # set l/r volumes
writespc("\x02\x11\x1f"); # set l/r volumes
writespc("\x02\x20\x1f"); # set l/r volumes
writespc("\x02\x21\x1f"); # set l/r volumes
writespc("\x02\x30\x1f"); # set l/r volumes
writespc("\x02\x31\x1f"); # set l/r volumes
writespc("\x02\x40\x1f"); # set l/r volumes
writespc("\x02\x41\x1f"); # set l/r volumes
writespc("\x02\x50\x1f"); # set l/r volumes
writespc("\x02\x51\x1f"); # set l/r volumes
writespc("\x02\x60\x1f"); # set l/r volumes
writespc("\x02\x61\x1f"); # set l/r volumes
writespc("\x02\x70\x1f"); # set l/r volumes
writespc("\x02\x71\x1f"); # set l/r volumes
writespc("\x06"); # RTS

print "pattern order table: ";
foreach $foo (@pattern_order) {
	print "$foo ";
}
print "\n";
print "restart position: $restartpos\n";
print "song length: $songlength\n";

# write the wrapper that inits the fu and calls all the patterns
$wrapper_addr=$spcaddr;
$braddr=calculate_branch($mss_initializer_addr);
writespc("\x05$braddr"); # BRS to init code
#foreach $foo (@pattern_order) {
#	if($foo<=$patterns) {
#		$braddr=calculate_branch($pattern[$foo]->{spcaddr});
#		writespc("\x05$braddr"); # BRS to this pattern
#	}
#}
$restartaddr=$wrapper_addr;
for($i=0;$i<$songlength;$i++) {
	$restartaddr=$spcaddr if($i==$restartpos);
	$foo=$pattern_order[$i];
	if($foo<=$patterns) {
		$braddr=calculate_branch($pattern[$foo]->{spcaddr});
		writespc("\x05$braddr"); # BRS to this pattern
	}
}
$braddr=calculate_branch($restartaddr);
writespc("\x04$braddr"); # BRA to pattern loop point
$braddr=calculate_branch($wrapper_addr);
writespc("\x04$braddr"); # kludge - BRA to wrapper beginning

# kludge to skip the patterns that come out bad
$wrapper_addr=$spcaddr;
$braddr=calculate_branch($mss_initializer_addr);
writespc("\x05$braddr"); # BRS to init code
$braddr=calculate_branch($pattern[0]->{spcaddr});
writespc("\x05$braddr"); # BRS to this pattern
#$restartaddr=$spcaddr;
$braddr=calculate_branch($pattern[1]->{spcaddr});
writespc("\x05$braddr"); # BRS to this pattern
$restartaddr=$spcaddr;
$braddr=calculate_branch($pattern[2]->{spcaddr});
writespc("\x05$braddr"); # BRS to this pattern
$braddr=calculate_branch($pattern[4]->{spcaddr});
writespc("\x05$braddr"); # BRS to this pattern
$braddr=calculate_branch($restartaddr);
writespc("\x04$braddr"); # BRA to pattern loop point
$braddr=calculate_branch($wrapper_addr);
writespc("\x04$braddr"); # kludge - BRA to wrapper beginning

# write the mss player to the file
close SPC;
system "dd of=level_2.apu if=SPCPLAY.BIN obs=512 seek=1 conv=notrunc";
print "Song entry point is $wrapper_addr\n";

sub readbyte {
	my $foo;

	read(XM,$foo,1) || die "read file: $!, died";
	return ord($foo);
}

sub readword {
	my $foo;

	read(XM,$foo,2) || die "read file: $!, died";
	return unpack('v',$foo);
}

sub readlong {
	my $foo;

	read(XM,$foo,4) || die "read file: $!, died";
	return unpack('V',$foo);
}

sub readnote {
# uses globals $data, $dataptr
	my %notedata;
	my $firstbyte;
	my $bits;

	%notedata=(note=>0, instrument=>0, volume=>0, effect_type=>0, effect_parm=>0);

	$firstbyte=ord(substr($data,$dataptr,1));
	$dataptr++;
	if($firstbyte&0x80) { # packed format
		$bits=$firstbyte;
	} else { # 5-byte format
		#$bits=~1;
		$bits=0xfe; # everything but the note
		$notedata{note}=$firstbyte;
	}

	printf "readnote: firstbyte=0x%x, bits=0x%x\n",$firstbyte,$bits;

	if($bits&1) {
		$notedata{note}=ord(substr($data,$dataptr,1));
		$dataptr++;
	}
	if($bits&2) {
		$notedata{instrument}=ord(substr($data,$dataptr,1));
		$dataptr++;
	}
	if($bits&4) {
		$notedata{volume}=ord(substr($data,$dataptr,1));
		$dataptr++;
	}
	if($bits&8) {
		$notedata{effect_type}=ord(substr($data,$dataptr,1));
		$dataptr++;
	}
	if($bits&16) {
		$notedata{effect_parm}=ord(substr($data,$dataptr,1));
		$dataptr++;
	}

	return %notedata;
}

sub writespc {
	my $data=shift;
	seek SPC,$spcaddr-0x200,0;
	print SPC $data;
	$spcaddr+=length($data);
}

sub writesample {
	my($instnum, $sampnum)=@_;
	my $loop;
	my $filter=0;

	my @pcmdata=decodesample($instnum, $sampnum);
	$loop=1 if($inst[$instnum]->{samp}->[$sampnum]->{looplength} != 0 && $inst[$instnum]->{samp}->[$sampnum]->{flags}&3 != 0);
	my $brrdata=encodesample($loop,$filter,@pcmdata);
	printf "writing sample. scalar(pcmdata)=%d, length(brrdata)=%d, spcaddr=0x%x\n",scalar(@pcmdata),length($brrdata),$spcaddr;
	writespc($brrdata);
	open FOO,">pcm.$instnum.$sampnum";
	foreach my $demon (@pcmdata) {
		print FOO chr($demon&0xff);
		print FOO chr(($demon>>8)&0xff);
	}
	close FOO;
}

sub decodesample {
	my($instnum, $sampnum)=@_;
	my @pcmdata;
	my $bits;
	my $samplecount;
	my $offset=0;
	my $thissample;
	my $lastsample=0;
	my $i;

	# the sample data in the xm can be 8 or 16 bits. in both cases, the data
	# is stored with full (8 or 16 bit) width, but is delta encoded. wtf?
	if($inst[$instnum]->{samp}->[$sampnum]->{flags}&0x10) {
		$bits=16;
	} else {
		$bits=8;
	}
	print "Decoding instrument $instnum sample $sampnum, $bits bits\n";

	$samplecount=$inst[$instnum]->{samp}->[$sampnum]->{length};
	$samplecount/=2 if($bits==16);

	#printf "%d $offset\n",length($inst[$instnum]->{samp}->[$sampnum]->{data});
	#for($xxx=0;$xxx<length($inst[$instnum]->{samp}->[$sampnum]->{data});$xxx++) {
	#	printf("%02x ",ord(substr($inst[$instnum]->{samp}->[$sampnum]->{data},$xxx,1)));
	#}
	#print "\n";
	for($i=0;$i<$samplecount;$i++) {
		$thissample=ord(substr($inst[$instnum]->{samp}->[$sampnum]->{data},$offset,1));
		#printf("%02x ",$thissample);
		$offset++;
		if($bits==16) {
			$thissample|=ord(substr($inst[$instnum]->{samp}->[$sampnum]->{data},$offset,1))<<8;
			$offset++;
		} else {
			$thissample<<=8;
		}
		$thissample+=$lastsample;
		$thissample&=0xffff;
		$lastsample=$thissample;
		#if($thissample>32767) {
		#	$thissample=32767-$thissample;
		#}
		#$thissample-=32767;
		#printf("%04x ",$thissample);
		#if($thissample>32767) {
		#	$thissample-=65536;
		#}
		push @pcmdata, $thissample;
	}

	return @pcmdata;
}

sub calculate_range_1samp {
	my $samp=shift;

	return	0	if($samp>=	-8	&& $samp<=	7	);
	return	1	if($samp>=	-16	&& $samp<=	14	);
	return	2	if($samp>=	-32	&& $samp<=	28	);
	return	3	if($samp>=	-64	&& $samp<=	56	);
	return	4	if($samp>=	-128	&& $samp<=	112	);
	return	5	if($samp>=	-256	&& $samp<=	224	);
	return	6	if($samp>=	-512	&& $samp<=	448	);
	return	7	if($samp>=	-1024	&& $samp<=	896	);
	return	8	if($samp>=	-2048	&& $samp<=	1792	);
	return	9	if($samp>=	-4096	&& $samp<=	3584	);
	return	10	if($samp>=	-8192	&& $samp<=	7168	);
	return	11	if($samp>=	-16384	&& $samp<=	14336	);
	return	12	if($samp>=	-32768	&& $samp<=	28672	);
	return 12;
}

sub calculate_range {
	my(@pcmblock)=@_;
	my $range=0;
	my $newrange;
	my $sample;

	foreach $sample (@pcmblock) {
		$newrange=calculate_range_1samp($sample);
		$range=$newrange if($newrange>$range);
	}

	return $range;
}

sub encodesample {
	my($loop,$filter,@pcmdata)=@_;
	my @pcmblock;
	my $brrdata;
	my $range;
	my $end=0;
	my $byte;
	my $i;
	my $sample;

	for(;;) {
		# grab 16 samples
		@pcmblock=splice(@pcmdata,0,16);
		while(scalar(@pcmblock)<16) { push @pcmblock, 0; } # pad
		$end++ if(scalar(@pcmdata)==0);
		# calculate the range
		$range=calculate_range(@pcmblock);
		# encode the header
		$byte=0;
		$byte|=1 if($end);
		$byte|=2 if($loop);
		$byte|=$filter<<2;
		$byte|=$range<<4;
		$brrdata.=chr($byte);
		# encode the samples
		for($i=0;$i<8;$i++) {
			$sample=shift @pcmblock;
			$byte=($sample>>$range)&0xf;
			$byte<<=4;
			$sample=shift @pcmblock;
			$byte|=($sample>>$range)&0xf;
			$brrdata.=chr($byte);
		}
		# done? return
		return $brrdata if($end);
	}
}

sub calculate_branch {
	my $destaddr=shift;
	my $srcaddr=$spcaddr+3;
	my $offset=($destaddr-$srcaddr)&0xffff;

	#printf "Branch from %x to %x, displacement is %x\n",$srcaddr,$destaddr,$offset;
	return(chr($offset&0xff).chr($offset>>8));
}

sub figure_ticks {
	# we want number of mss clock ticks (1/128 of a sec) per division (note)
	my $divisions_per_minute=(24*$default_bpm)/$default_tempo;
	my $seconds_per_division=60/$divisions_per_minute;
	my $mssticks_per_division=128*$seconds_per_division;
	#$mssticks_per_division--; # kludge for spc code's slowness
	print "default_tempo=$default_tempo, default_bpm=$default_bpm\n";
	print "$divisions_per_minute divisions/minute, $seconds_per_division seconds/division, =$mssticks_per_division mssticks/division\n";
	return int($mssticks_per_division);
}

sub figure_srcn {
	my($instnum,$note)=@_;
	my $samplenum;
	my $srcn;

	$instnum--;
	$note--;
	$samplenum=$inst[$instnum]->{samplenum}->[$note];
	$srcn=$inst[$instnum]->{samp}->[$samplenum]->{apusample};
#	return 2;
	return $srcn;
}

sub xm_player_get_note_pitch {
# (soundtracker xm-player.c)
	my $procnot=shift; # this is the note value straight from the pattern
	my $relnote=shift; # from sample, with sign processed
	my $finetune=shift; # from sample, with sign processed
	my $curnormnote=-$relnote*256 - $finetune*2;
	my $pitch = 48*256 - ((($procnot - 1) << 8) - $curnormnote);
	return $pitch if($freqflag&1);
	die "don't grok amiga pitch tables, died";
}

sub pitch_to_freq {
	my $pitch=shift;
	return 8363 * 2**((-$pitch)/(12*16*4*4));
}

sub figure_pitch {
	my($instnum,$note)=@_;
	my $samplenum;
	my $realnote;
	my $period;
	my $finetune;
	my $hz;
	my $factor;
	my $pitchval;
	my $pitch;

	$instnum--;
	$samplenum=$inst[$instnum]->{samplenum}->[$note-1];

	$pitch=xm_player_get_note_pitch($note,$inst[$instnum]->{samp}->[$samplenum]->{relnote},$inst[$instnum]->{samp}->[$samplenum]->{finetune}); # xm pitch number, not spc pitch number
	$hz=pitch_to_freq($pitch);

	print "frequency $hz hz\n";
	# ok, i guess we have the "absolute" playback rate in Hz... now
	# normalize it to the apu's 32k sample rate
	$factor=$hz/32000;
	# now we should have e.g. $factor=2 for a 64k playback rate, or
	# $factor=0.5 for a 16k playback rate
	$pitchval=int($factor*4096);
	print "pitchval=$pitchval\n";
	die "pitch $pitchval is too high" if(($pitchval&0xC000)!=0);
	return(chr($pitchval&0xff).chr($pitchval>>8));
	#return(chr($pitchval>>8).chr($pitchval&0xff));
}
